# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
import subprocess
from pathlib import Path
from textwrap import dedent
from typing import Callable, Iterable

import black

logger = logging.getLogger(__name__)


def sanitize_name(raw, expect_arithmetic: bool = False):
    sanitized = raw
    # braces
    sanitized = sanitized.replace("(", "")
    sanitized = sanitized.replace(")", "")
    sanitized = sanitized.replace("[", "")
    sanitized = sanitized.replace("]", "")
    # seperators
    sanitized = sanitized.replace(".", "_")
    sanitized = sanitized.replace(",", "_")
    sanitized = sanitized.replace("/", "_")
    if not expect_arithmetic:
        sanitized = sanitized.replace("-", "_")

    # special symbols
    sanitized = sanitized.replace("'", "")
    sanitized = sanitized.replace("*", "")
    sanitized = sanitized.replace("^", "p")
    sanitized = sanitized.replace("#", "h")
    sanitized = sanitized.replace("ϕ", "phase")
    sanitized = sanitized.replace("π", "pi")
    sanitized = sanitized.replace("&", "and")
    # inversion
    sanitized = sanitized.replace("~", "n")
    sanitized = sanitized.replace("{", "")
    sanitized = sanitized.replace("}", "")

    sanitized = sanitized.replace("->", "to")
    sanitized = sanitized.replace("<-", "from")
    # arithmetics
    sanitized = sanitized.replace(">", "gt")
    sanitized = sanitized.replace("<", "lt")
    sanitized = sanitized.replace("=", "eq")
    sanitized = sanitized.replace("+", "plus")
    sanitized = sanitized.replace("-", "minus")

    # rest
    def handle_unknown_invalid_symbold(match):
        logger.warning(
            'Replacing unknown invalid symbol "{}" in "{}" with _'.format(
                match.group(0), raw
            )
        )
        return "_"

    sanitized = re.sub(r"[^a-zA-Z_0-9]", handle_unknown_invalid_symbold, sanitized)

    if re.match("^[a-zA-Z_]", sanitized) is None:
        sanitized = "P" + sanitized

    if re.match("^[a-zA-Z_]+[a-zA-Z_0-9]*$", sanitized) is not None:
        return sanitized

    to_escape = re.findall("[^a-zA-Z_0-9]", sanitized)
    if len(to_escape) > 0:
        raise ValueError(f"Cannot sanitize name: {raw}")

    return sanitized


def gen_repeated_block[T](
    generator: Iterable[T],
    func: Callable[[T], str] = dedent,
    requires_pass: bool = False,
) -> str:
    lines = list(map(func, generator))

    if not lines and requires_pass:
        lines = ["pass"]

    return gen_block("\n".join(lines))


def gen_block(payload: str):
    return f"#__MARK_BLOCK_BEGIN\n{payload}\n#__MARK_BLOCK_END"


def fix_indent(text: str) -> str:
    indent_stack = [""]

    out_lines = []
    for line in text.splitlines():
        if "#__MARK_BLOCK_BEGIN" in line:
            indent_stack.append(line.removesuffix("#__MARK_BLOCK_BEGIN"))
        elif "#__MARK_BLOCK_END" in line:
            indent_stack.pop()
        else:
            out_lines.append(indent_stack[-1] + line)

    return dedent("\n".join(out_lines))


def format_and_write(code: str, path: Path):
    code = code.strip()
    code = black.format_file_contents(code, fast=True, mode=black.FileMode())
    path.write_text(code, encoding="utf-8")

    subprocess.check_output(["ruff", "check", "--fix", path])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class _BH1750FVI_TR(Module):
    class _bh1750_esphome_config(F.has_esphome_config.impl()):
        update_interval = L.p_field(
            units=P.s,
            soft_set=L.Range(100 * P.ms, 1 * P.day),
            guess=1 * P.s,
            tolerance_guess=0,
        )

        def get_config(self) -> dict:
            obj = self.obj
            assert isinstance(obj, _BH1750FVI_TR)

            i2c = F.is_esphome_bus.find_connected_bus(obj.i2c)

            return {
                "sensor": [
                    {
                        "platform": "bh1750",
                        "name": "BH1750 Illuminance",
                        "address": "0x23",
                        "i2c_id": i2c.get_trait(F.is_esphome_bus).get_bus_id(),
                        "update_interval": self.update_interval,
                    }
                ]
            }

    dvi_capacitor: F.Capacitor
    dvi_resistor: F.Resistor

    power: F.ElectricPower
    addr: F.ElectricLogic
    dvi: F.ElectricLogic
    ep: F.ElectricLogic
    i2c: F.I2C

    def set_address(self, addr: int):
        raise NotImplementedError()
        # TODO: Implement set_address
        # ADDR = ‘H’ ( ADDR ≧ 0.7VCC ) “1011100“
        # ADDR = 'L' ( ADDR ≦ 0.3VCC ) “0100011“
        ...
        # assert addr < (1 << len(self.e))

        # for i, e in enumerate(self.e):
        #    e.set(addr & (1 << i) != 0)

    esphome_config: _bh1750_esphome_config

    def __preinit__(self):
        self.dvi_capacitor.capacitance.constrain_subset(
            L.Range.from_center_rel(1 * P.uF, 0.1)
        )
        self.dvi_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1 * P.kohm, 0.1)
        )

        self.i2c.frequency.constrain_le(
            F.I2C.define_max_frequency_capability(F.I2C.SpeedMode.fast_speed)
        )

        # set constraints
        self.power.voltage.constrain_subset(L.Range(2.4 * P.V, 3.6 * P.V))

        # TODO: self.dvi.low_pass(self.dvi_capacitor, self.dvi_resistor)
        self.dvi.line.connect_via(self.dvi_capacitor, self.power.lv)
        self.dvi.line.connect_via(self.dvi_resistor, self.power.hv)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power.hv,
                "2": self.addr.line,
                "3": self.power.lv,
                "4": self.i2c.sda.line,
                "5": self.dvi.line,
                "6": self.i2c.scl.line,
                "7": self.ep.line,
            }
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheet.lcsc.com/lcsc/1811081611_ROHM-Semicon-BH1750FVI-TR_C78960.pdf"
    )

    mfr = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "ROHM Semicon",
            DescriptiveProperties.partno: "BH1750FVI-TR",
        }
    )

    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C78960"})


# TODO should be a reference design
class BH1750FVI_TR(Module):
    ic: _BH1750FVI_TR

    def __preinit__(self):
        self.ic.i2c.terminate(self)

        self.ic.power.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.1)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import pprint
from copy import deepcopy
from functools import singledispatch
from itertools import chain, groupby
from os import PathLike
from pathlib import Path
from typing import Any, List, Protocol

# import numpy as np
# from shapely import Polygon
import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.libs.exceptions import UserException
from faebryk.libs.geometry.basic import Geometry
from faebryk.libs.kicad.fileformats_common import C_effects, C_pts, C_wh, C_xy, C_xyr
from faebryk.libs.kicad.fileformats_common import (
    gen_uuid as _gen_uuid,
)
from faebryk.libs.kicad.fileformats_latest import (
    C_kicad_fp_lib_table_file,
)
from faebryk.libs.kicad.fileformats_sch import (
    UUID,
    C_arc,
    C_circle,
    C_kicad_sch_file,
    C_kicad_sym_file,
    C_polyline,
    C_property,
    C_rect,
    C_stroke,
)
from faebryk.libs.kicad.paths import GLOBAL_FP_DIR_PATH, GLOBAL_FP_LIB_PATH
from faebryk.libs.sexp.dataclass_sexp import dataclass_dfs
from faebryk.libs.util import (
    cast_assert,
    find,
    get_key,
    not_none,
    once,
)

logger = logging.getLogger(__name__)


SCH = C_kicad_sch_file.C_kicad_sch

Geom = C_polyline | C_arc | C_rect | C_circle
Font = C_effects.C_font

Point = Geometry.Point
Point2D = Geometry.Point2D

Justify = C_effects.C_justify.E_justify
Alignment = tuple[Justify, Justify, Justify]
Alignment_Default = (Justify.center_horizontal, Justify.center_vertical, Justify.normal)


def gen_uuid(mark: str = "") -> UUID:
    return _gen_uuid(mark)


def is_marked(uuid: UUID, mark: str):
    suffix = mark.encode().hex()
    return uuid.replace("-", "").endswith(suffix)


class _HasUUID(Protocol):
    uuid: UUID


# TODO: consider common transformer base
class SchTransformer:
    class has_linked_sch_symbol(Module.TraitT):
        symbol: SCH.C_symbol_instance

    class has_linked_sch_symbol_defined(has_linked_sch_symbol.impl()):
        def __init__(self, symbol: SCH.C_symbol_instance) -> None:
            super().__init__()
            self.symbol = symbol

    class has_linked_pins(F.Symbol.Pin.TraitT):
        pins: list[SCH.C_symbol_instance.C_pin]

    class has_linked_pins_defined(has_linked_pins.impl()):
        def __init__(
            self,
            pins: list[SCH.C_symbol_instance.C_pin],
        ) -> None:
            super().__init__()
            self.pins = pins

    def __init__(
        self, sch: SCH, graph: Graph, app: Module, cleanup: bool = True
    ) -> None:
        self.sch = sch
        self.graph = graph
        self.app = app
        self._symbol_files_index: dict[str, Path] = {}

        self.missing_lib_symbols: list[SCH.C_lib_symbols.C_symbol] = []

        self.dimensions = None

        FONT_SCALE = 8
        FONT = Font(
            size=C_wh(1 / FONT_SCALE, 1 / FONT_SCALE),
            thickness=0.15 / FONT_SCALE,
        )
        self.font = FONT

        # TODO: figure out what to do with cleanup
        # self.cleanup()
        self.attach()

    def attach(self):
        """This function matches and binds symbols to their symbols"""
        # reference (eg. C3) to symbol (eg. "Capacitor_SMD:C_0402")
        symbols = {
            (f.propertys["Reference"].value, f.lib_id): f for f in self.sch.symbols
        }
        for node, sym_trait in GraphFunctions(self.graph).nodes_with_trait(
            F.Symbol.has_symbol
        ):
            # FIXME: I believe this trait is used as a proxy for being a component
            # since, names are replaced with designators during typical pipelines
            if not node.has_trait(F.has_overriden_name):
                continue

            symbol = sym_trait.reference

            if not symbol.has_trait(F.Symbol.has_kicad_symbol):
                continue

            sym_ref = node.get_trait(F.has_overriden_name).get_name()
            sym_name = symbol.get_trait(F.Symbol.has_kicad_symbol).symbol_name

            try:
                sym = symbols[(sym_ref, sym_name)]
            except KeyError:
                # TODO: add diag
                self.missing_lib_symbols.append(symbol)
                continue

            self.attach_symbol(node, sym)

        # Log what we were able to attach
        attached = {
            n: t.symbol
            for n, t in GraphFunctions(self.graph).nodes_with_trait(
                SchTransformer.has_linked_sch_symbol
            )
        }
        logger.debug(f"Attached: {pprint.pformat(attached)}")

        if self.missing_lib_symbols:
            # TODO: just go look for the symbols instead
            raise ExceptionGroup(
                "Missing lib symbols",
                [
                    f"Symbol {sym.name} not found in symbols dictionary"
                    for sym in self.missing_lib_symbols
                ],
            )

    def attach_symbol(self, node: Node, symbol: SCH.C_symbol_instance):
        """Bind the module and symbol together on the graph"""
        graph_sym = node.get_trait(F.Symbol.has_symbol).reference

        graph_sym.add(self.has_linked_sch_symbol_defined(symbol))

        # Attach the pins on the symbol to the module interface
        for pin_name, pins in groupby(symbol.pins, key=lambda p: p.name):
            graph_sym.pins[pin_name].add(SchTransformer.has_linked_pins_defined(pins))

    def cleanup(self):
        """Delete faebryk-created objects in schematic."""

        # find all objects with path_len 2 (direct children of a list in pcb)
        candidates = [o for o in dataclass_dfs(self.sch) if len(o[1]) == 2]
        for obj, path, _ in candidates:
            if not self.is_marked(obj):
                continue

            # delete object by removing it from the container they are in
            holder = path[-1]
            if isinstance(holder, list):
                holder.remove(obj)
            elif isinstance(holder, dict):
                del holder[get_key(obj, holder)]

    def index_symbol_files(
        self, fp_lib_tables: PathLike | list[PathLike], load_globals: bool = True
    ) -> None:
        if isinstance(fp_lib_tables, (str, Path)):
            fp_lib_table_paths = [Path(fp_lib_tables)]
        else:
            fp_lib_table_paths = [Path(p) for p in fp_lib_tables]

        # non-local lib, search in kicad global lib
        if load_globals:
            fp_lib_table_paths += [GLOBAL_FP_LIB_PATH]

        for lib_path in fp_lib_table_paths:
            for lib in C_kicad_fp_lib_table_file.loads(lib_path).fp_lib_table.libs:
                resolved_lib_dir = Path(
                    (lib.uri)
                    .replace("${KIPRJMOD}", str(lib_path.parent))
                    .replace("${KICAD9_FOOTPRINT_DIR}", str(GLOBAL_FP_DIR_PATH))
                )
                for path in resolved_lib_dir.glob("*.kicad_sym"):
                    if path.stem not in self._symbol_files_index:
                        self._symbol_files_index[path.stem] = path

    @staticmethod
    def flipped[T](input_list: list[tuple[T, int]]) -> list[tuple[T, int]]:
        return [(x, (y + 180) % 360) for x, y in reversed(input_list)]

    @staticmethod
    def gen_uuid(mark: bool = False):
        return gen_uuid(mark="FBRK" if mark else "")

    @staticmethod
    def is_marked(obj) -> bool:
        if not hasattr(obj, "uuid"):
            return False
        return is_marked(obj.uuid, "FBRK")

    # Getter ---------------------------------------------------------------------------
    @staticmethod
    def get_symbol(cmp: Node) -> F.Symbol:
        return not_none(cmp.get_trait(SchTransformer.has_linked_sch_symbol)).symbol

    def get_all_symbols(self) -> List[tuple[Module, F.Symbol]]:
        return [
            (cast_assert(Module, cmp), t.symbol)
            for cmp, t in GraphFunctions(self.graph).nodes_with_trait(
                SchTransformer.has_linked_sch_symbol
            )
        ]

    @once
    def get_symbol_file(self, lib_name: str) -> C_kicad_sym_file:
        # primary caching handled by @once
        if lib_name not in self._symbol_files_index:
            raise UserException(f"Symbol file {lib_name} not found")

        path = self._symbol_files_index[lib_name]
        return C_kicad_sym_file.loads(path)

    @staticmethod
    def get_related_lib_sym_units(
        lib_sym: SCH.C_lib_symbols.C_symbol,
    ) -> dict[int, list[SCH.C_lib_symbols.C_symbol.C_symbol]]:
        """
        Figure out units.
        This seems to be purely based on naming convention.
        There are two suffixed numbers on the end eg. _0_0, _0_1
        They're in two sets of groups:
            1. subunit. used to represent graphical vs. pin objects within a unit
            2. unit. eg, a single op-amp in a package with 4
        We need to lump the subunits together for further processing.

        That is, we group them by the last number.
        """
        groups = groupby(lib_sym.symbols.items(), key=lambda item: int(item[0][-1]))
        return {k: [v[1] for v in vs] for k, vs in groups}

    @singledispatch
    def get_lib_symbol(self, sym) -> SCH.C_lib_symbols.C_symbol:
        raise NotImplementedError(f"Don't know how to get lib symbol for {type(sym)}")

    @get_lib_symbol.register
    def _(self, sym: F.Symbol) -> SCH.C_lib_symbols.C_symbol:
        lib_id = sym.get_trait(F.Symbol.has_kicad_symbol).symbol_name
        return self._ensure_lib_symbol(lib_id)

    @get_lib_symbol.register
    def _(self, sym: SCH.C_symbol_instance) -> SCH.C_lib_symbols.C_symbol:
        return self.sch.lib_symbols.symbols[sym.lib_id]

    @singledispatch
    def get_lib_pin(self, pin) -> SCH.C_lib_symbols.C_symbol.C_symbol.C_pin:
        raise NotImplementedError(f"Don't know how to get lib pin for {type(pin)}")

    @get_lib_pin.register
    def _(self, pin: F.Symbol.Pin) -> SCH.C_lib_symbols.C_symbol.C_symbol.C_pin:
        graph_symbol, _ = pin.get_parent()
        assert isinstance(graph_symbol, Node)
        lib_sym = self.get_lib_symbol(graph_symbol)
        units = self.get_related_lib_sym_units(lib_sym)
        sym = graph_symbol.get_trait(SchTransformer.has_linked_sch_symbol).symbol

        def _name_filter(sch_pin: SCH.C_lib_symbols.C_symbol.C_symbol.C_pin):
            return sch_pin.name in {
                p.name for p in pin.get_trait(self.has_linked_pins).pins
            }

        lib_pin = find(
            chain.from_iterable(u.pins for u in units[sym.unit]),
            _name_filter,
        )
        return lib_pin

    # Insert ---------------------------------------------------------------------------
    @staticmethod
    def mark[R: _HasUUID](node: R) -> R:
        if hasattr(node, "uuid"):
            node.uuid = SchTransformer.gen_uuid(mark=True)  # type: ignore

        return node

    def _get_list_field[R](self, node: R, prefix: str = "") -> list[R]:
        root = self.sch
        key = prefix + type(node).__name__.removeprefix("C_") + "s"

        assert hasattr(root, key)

        target = getattr(root, key)
        assert isinstance(target, list)
        assert all(isinstance(x, type(node)) for x in target)
        return target

    def _insert(self, obj: Any, prefix: str = ""):
        obj = SchTransformer.mark(obj)
        self._get_list_field(obj, prefix=prefix).append(obj)

    def _delete(self, obj: Any, prefix: str = ""):
        self._get_list_field(obj, prefix=prefix).remove(obj)

    def insert_wire(
        self,
        coords: list[Geometry.Point2D],
        stroke: C_stroke | None = None,
    ):
        """Insert a wire with points at all the coords"""
        for section in zip(coords[:-1], coords[1:]):
            self.sch.wires.append(
                SCH.C_wire(
                    pts=C_pts(xys=[C_xy(*coord) for coord in section]),
                    stroke=stroke or C_stroke(),
                )
            )

    def insert_text(
        self,
        text: str,
        at: C_xyr,
        font: Font,
        alignment: Alignment | None = None,
    ):
        self.sch.texts.append(
            SCH.C_text(
                text=text,
                at=at,
                effects=C_effects(
                    font=font,
                    justify=alignment,
                ),
                uuid=self.gen_uuid(mark=True),
            )
        )

    def _ensure_lib_symbol(
        self,
        lib_id: str,
    ) -> SCH.C_lib_symbols.C_symbol:
        """Ensure a symbol is in the schematic library, and return it"""
        if lib_id in self.sch.lib_symbols.symbols:
            return self.sch.lib_symbols.symbols[lib_id]

        lib_name, symbol_name = lib_id.split(":")
        lib_sym = deepcopy(
            self.get_symbol_file(lib_name).kicad_symbol_lib.symbols[symbol_name]
        )
        lib_sym.name = lib_id
        self.sch.lib_symbols.symbols[lib_id] = lib_sym
        return lib_sym

    def insert_symbol(
        self,
        module: Module,
        at: Point2D | None = None,
        rotation: int | None = None,
    ):
        if at is None:
            at = (0, 0)

        if rotation is None:
            rotation = 0

        # Symbols are attached to modules earlier in the pipeline
        # Typically, by the picker (lcsc.py), but plausibly by a library component
        symbol = module.get_trait(F.Symbol.has_symbol).reference

        # Ensure lib symbol is in sch
        lib_id = symbol.get_trait(F.Symbol.has_kicad_symbol).symbol_name
        lib_sym = self._ensure_lib_symbol(lib_id)

        # insert all units
        for unit_key, unit_objs in self.get_related_lib_sym_units(lib_sym).items():
            pins = []

            for subunit in unit_objs:
                for pin in subunit.pins:
                    pins.append(
                        SCH.C_symbol_instance.C_pin(
                            name=pin.name.name,
                            uuid=self.gen_uuid(mark=True),
                        )
                    )

            unit_instance = SCH.C_symbol_instance(
                lib_id=lib_id,
                unit=unit_key + 1,  # yes, these are indexed from 1...
                at=C_xyr(at[0], at[1], rotation),
                in_bom=True,
                on_board=True,
                pins=pins,
                uuid=self.gen_uuid(mark=True),
            )

            # Add a C_property for the reference based on the override name
            if reference_name := module.get_trait(F.has_overriden_name).get_name():
                unit_instance.propertys["Reference"] = C_property(
                    name="Reference",
                    value=reference_name,
                )
            else:
                # TODO: handle not having an overriden name better
                raise Exception(f"Module {module} has no overriden name")

            self.attach_symbol(module, unit_instance)

            self.sch.symbols.append(unit_instance)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class CH342F_ReferenceDesign(Module):
    """
    Minimal reference implementation of the CH342F.

    - Single power source (USB)
    - IO at 3.3V
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb_uart_converter: F.CH342F

    def __preinit__(self):
        # ----------------------------------------
        #                aliasess
        # ----------------------------------------
        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        self.usb_uart_converter.set_power_configuration(
            F.CH342.ChipPowerConfiguration.USB_5V,
            F.CH342.IOPowerConfiguration.INTERNAL_3V3,
        )

        # ----------------------------------------
        #              connections
        # ----------------------------------------
        self.usb_uart_converter.integrated_regulator.power_in.get_trait(
            F.can_be_decoupled
        ).decouple(owner=self)
        self.usb_uart_converter.power_3v.get_trait(F.can_be_decoupled).decouple(
            owner=self
        )
        self.usb_uart_converter.power_io.get_trait(F.can_be_decoupled).decouple(
            owner=self
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module  # noqa: F401
from faebryk.libs.library import L  # noqa: F401


class LEDIndicator(Module):
    # interfaces

    logic_in: F.ElectricLogic
    power_in: F.ElectricPower

    # components

    led: F.PoweredLED

    def __init__(self, use_mosfet: bool = False, active_low: bool = False):
        super().__init__()
        self._use_mosfet = use_mosfet
        self._active_low = active_low

    def __preinit__(self):
        if self._use_mosfet:
            power_switch = self.add(
                F.PowerSwitchMOSFET(lowside=True, normally_closed=self._active_low)
            )
        else:
            power_switch = self.add(
                F.PowerSwitchStatic(normally_closed=self._active_low)
            )

        self.power_in.connect_via(power_switch, self.led.power)
        power_switch.logic_in.connect(self.logic_in)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
from pathlib import Path
from typing import Sequence

from atopile.config import config
from faebryk.libs.exceptions import (
    UserResourceException,
    accumulate,
)
from faebryk.libs.kicad.fileformats_latest import (
    C_footprint,
    C_kicad_fp_lib_table_file,
)
from faebryk.libs.kicad.fileformats_version import kicad_footprint_file
from faebryk.libs.kicad.paths import GLOBAL_FP_DIR_PATH, GLOBAL_FP_LIB_PATH
from faebryk.libs.util import KeyErrorNotFound, find

logger = logging.getLogger(__name__)


class LibNotInTable(Exception):
    def __init__(self, *args: object, lib_id: str, lib_table_path: Path) -> None:
        super().__init__(*args)
        self.lib_id = lib_id
        self.lib_table_path = lib_table_path


def _find_footprint(
    lib_tables: Sequence[os.PathLike], lib_id: str
) -> C_kicad_fp_lib_table_file.C_fp_lib_table.C_lib:
    lib_tables = [Path(lib_table) for lib_table in lib_tables]

    err_accumulator = accumulate(LibNotInTable, FileNotFoundError)

    for lib_table_path in lib_tables:
        with err_accumulator.collect():
            lib_table = C_kicad_fp_lib_table_file.loads(lib_table_path)
            try:
                return find(lib_table.fp_lib_table.libs, lambda x: x.name == lib_id)
            except KeyErrorNotFound as ex:
                raise LibNotInTable(
                    lib_id=lib_id, lib_table_path=lib_table_path
                ) from ex

    if ex := err_accumulator.get_exception():
        raise ex

    raise ValueError("No footprint libraries provided")


def get_footprint(identifier: str, fp_lib_path: Path) -> C_footprint:
    lib_id, fp_name = identifier.split(":")
    lib = _find_footprint([fp_lib_path, GLOBAL_FP_LIB_PATH], lib_id)
    dir_path = Path(
        (lib.uri)
        .replace("${KIPRJMOD}", str(config.build.paths.fp_lib_table.parent))
        .replace("${KICAD9_FOOTPRINT_DIR}", str(GLOBAL_FP_DIR_PATH))
    )

    path = dir_path / f"{fp_name}.kicad_mod"
    try:
        return kicad_footprint_file(path).footprint
    except FileNotFoundError as ex:
        raise UserResourceException(
            f"Footprint `{fp_name}` doesn't exist in library `{lib_id}`"
        ) from ex



import logging
from pathlib import Path

import pcbnew  # type: ignore

from .common import (
    calculate_translation,
    flip_dict,
    generate_net_map,
    get_group_footprints,
    get_layout_map,
    log_exceptions,
    sync_drawing,
    sync_footprints,
    sync_track,
    sync_zone,
    update_zone_net,
)

log = logging.getLogger(__name__)


class PullGroup(pcbnew.ActionPlugin):
    def defaults(self):
        self.name = "Pull Group"
        self.category = "Pull Group Layout Atopile"
        self.description = (
            "Layout components on PCB in same spatial"
            " relationships as components on schematic"
        )
        self.show_toolbar_button = True
        self.icon_file_name = str(Path(__file__).parent / "download.png")
        self.dark_icon_file_name = self.icon_file_name

    @log_exceptions()
    def Run(self):
        target_board: pcbnew.BOARD = pcbnew.GetBoard()
        board_path = target_board.GetFileName()
        known_layouts = get_layout_map(board_path)

        combined_addr_map = {}
        combined_uuid_map = {}
        for group_key, group_val in known_layouts.items():
            # Expect each group value to be a dict with a 'layout_path' key
            if isinstance(group_val, dict) and "layout_path" in group_val:
                combined_addr_map.update(group_val.get("addr_map", {}))
                combined_uuid_map.update(group_val.get("uuid_to_addr_map", {}))

        log.info(f"Combined addr_map: {combined_addr_map}")
        log.info(f"Combined uuid_to_addr_map: {combined_uuid_map}")

        # Pull Selected Groups
        for g in target_board.Groups():
            assert isinstance(g, pcbnew.PCB_GROUP)

            # Skip over unselected groups
            if not g.IsSelected():
                continue

            g_name = g.GetName()
            if g_name not in known_layouts:
                continue

            layout_path = Path(known_layouts[g_name]["layout_path"])
            # Check what layouts exist
            if not layout_path.exists():
                raise FileNotFoundError(
                    f"Cannot load group. Layout file {layout_path} does not exist"
                )

            # Remove everything but the footprints from the target group
            for item in g.GetItems():
                if not isinstance(item, pcbnew.FOOTPRINT):
                    target_board.Remove(item)

            # Load the layout and sync
            source_board: pcbnew.BOARD = pcbnew.LoadBoard(str(layout_path))

            # Calculate offset before moving footprints
            offset = calculate_translation(
                source_fps=source_board.GetFootprints(),
                target_fps=get_group_footprints(g),
            )

            layout_maps = known_layouts[g_name]

            # Get addr_map and uuid_map for this specific layout
            layout_addr_map = layout_maps.get("addr_map", {})
            layout_uuid_map = layout_maps.get("uuid_to_addr_map", {})

            if "addr_map" not in layout_maps:
                log.error(
                    f"[DEBUG] layout_maps keys for group {g_name}: "
                    f"{list(layout_maps.keys())}"
                )
                log.warning(
                    f"'addr_map' key not found in layout_maps for group {g_name}; "
                    f"defaulting to empty mapping"
                )
            if "uuid_to_addr_map" not in layout_maps:
                log.error(
                    f"[DEBUG] layout_maps keys for group {g_name}: "
                    f"{list(layout_maps.keys())}"
                )
                log.warning(
                    f"'uuid_to_addr_map' key not found in layout_maps for"
                    f" group {g_name}; defaulting to empty mapping"
                )

            # Generate net_map specifically for this source layout and target board
            try:
                net_map = generate_net_map(
                    source_board=source_board,
                    target_board=target_board,
                    addr_map=flip_dict(layout_addr_map),
                    uuid_map=layout_uuid_map,
                )

                # Log some info about the nets for debugging
                # NetsByName() returns a dict where keys are the net names
                source_nets = list(source_board.GetNetInfo().NetsByName().keys())
                # Convert wxString to Python string for logging - used for display below
                source_nets_as_str = [str(net) for net in source_nets]
                log.info(f"Source board nets count: {len(source_nets)}")
                log.info(f"Net map entries count: {len(net_map)}")

                # Find unmapped nets that aren't empty
                unmapped_nets = [
                    net
                    for net in source_nets_as_str
                    if net not in net_map and net != ""
                ]
                log.info(f"Unmapped nets: {unmapped_nets}")

            except KeyError as e:
                log.error(
                    f"KeyError in generate_net_map for group {g_name}: {e}. "
                    f"Layout map keys: {list(layout_addr_map.keys())}"
                )
                raise

            sync_footprints(
                source_board,
                target_board,
                flip_dict(layout_maps.get("addr_map", {})),
                layout_maps.get("uuid_to_addr_map", {}),
            )

            for track in source_board.GetTracks():
                item = sync_track(source_board, track, target_board, net_map)
                g.AddItem(item)

            for drawing in source_board.GetDrawings():
                item = sync_drawing(source_board, drawing, target_board)
                g.AddItem(item)

            for zone in source_board.Zones():
                new_zone = sync_zone(zone, target_board)
                update_zone_net(
                    zone,
                    new_zone,
                    net_map,
                )
                g.AddItem(new_zone)

            # Shift entire target group by offset as last operation
            g.Move(offset)

        pcbnew.Refresh()


with log_exceptions():
    PullGroup().register()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class QWIIC_Connector(Module):
    """
    SparkFun's Qwiic Connect System uses 4-pin JST connectors to quickly interface
    development boards with sensors, LCDs, relays and more via I2C.
    1x4P 4P JST SH 1mm pitch horizontal mount
    """

    power: F.ElectricPower
    i2c: F.I2C

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.X
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "JST Sales America",
            DescriptiveProperties.partno: "SM04B-SRSS-TB(LF)(SN)",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)("https://www.sparkfun.com/qwiic")

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power.lv: ["1"],
                self.power.lv: ["2"],
                self.i2c.sda.line: ["3"],
                self.i2c.scl.line: ["4"],
                # n.c. ["5"],
                # n.c. ["6"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power.lv,
                "2": self.power.hv,
                "3": self.i2c.sda.line,
                "4": self.i2c.scl.line,
            }
        )

    def __preinit__(self):
        self.power.voltage.constrain_subset(L.Range.from_center(3.3 * P.V, 0.3 * P.V))
        # self.power.max_current.constrain_subset(
        #    L.Range.from_center_rel(226 * P.mA, 0.05)
        # )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.parameter import Parameter


class is_pickable_by_type(F.is_pickable.decless()):
    class Type(Enum):
        Resistor = auto()
        Capacitor = auto()
        Inductor = auto()
        TVS = auto()
        LED = auto()
        Diode = auto()
        LDO = auto()
        MOSFET = auto()

    def __init__(self, pick_type: Type, parameters: dict[str, Parameter]):
        super().__init__()
        self._pick_type = pick_type
        self._parameters = parameters

    def get_pick_type(self) -> Type:
        return self._pick_type

    def get_parameters(self) -> dict[str, Parameter]:
        return self._parameters



"""
This module contains a dataclass representation of the KiCad file formats
for schematic files (.kicad_sch).

Rules:
- Defaults are only allowed only with the values KiCAD itself defaults to
"""

import logging
from dataclasses import dataclass, field
from enum import auto
from typing import Optional

from faebryk.exporters.pcb.kicad.transformer import gen_uuid
from faebryk.libs.kicad.fileformats_common import (
    UUID,
    C_effects,
    C_pts,
    C_xy,
    C_xyr,
)
from faebryk.libs.sexp.dataclass_sexp import SEXP_File, SymEnum, sexp_field

logger = logging.getLogger(__name__)

# TODO find complete examples of the fileformats, maybe in the kicad repo


def uuid_field():
    return field(default_factory=gen_uuid)


@dataclass
class C_property:
    name: str = field(**sexp_field(positional=True))
    value: str = field(**sexp_field(positional=True))
    id: Optional[int] = None
    at: Optional[C_xyr] = None
    effects: Optional[C_effects] = field(
        **sexp_field(preprocessor=C_effects.preprocess_shitty_hide), default=None
    )


@dataclass(kw_only=True)  # TODO: when to use kw_only?
class C_fill:
    class E_type(SymEnum):
        background = "background"
        none = "none"
        outline = "outline"

    type: E_type = field(default=E_type.background)


@dataclass
class C_stroke:
    """
    KiCAD default: (stroke (width 0) (type default) (color 0 0 0 0))
    """

    class E_type(SymEnum):
        solid = auto()
        default = auto()

    width: float = 0
    type: E_type = E_type.default
    color: tuple[int, int, int, int] = (0, 0, 0, 0)


@dataclass(kw_only=True)
class C_circle:
    center: C_xy
    end: C_xy
    stroke: C_stroke
    fill: C_fill


@dataclass(kw_only=True)
class C_arc:
    start: C_xy
    mid: C_xy
    end: C_xy
    stroke: C_stroke
    fill: C_fill


@dataclass(kw_only=True)
class C_rect:
    start: C_xy
    end: C_xy
    stroke: C_stroke
    fill: C_fill


@dataclass(kw_only=True)
class C_polyline:
    stroke: C_stroke
    fill: C_fill
    pts: C_pts = field(default_factory=C_pts)


@dataclass
class C_kicad_sch_file(SEXP_File):
    """
    When in doubt check: kicad/eeschema/sch_io/kicad_sexpr/sch_io_kicad_sexpr_parser.cpp
    """

    @dataclass
    class C_kicad_sch:
        @dataclass
        class C_title_block:
            title: Optional[str] = None
            date: Optional[str] = None
            rev: Optional[str] = None
            company: Optional[str] = None

        @dataclass
        class C_lib_symbols:
            @dataclass
            class C_symbol:
                @dataclass
                class C_pin_names:
                    offset: float

                @dataclass
                class C_symbol:
                    @dataclass
                    class C_pin:
                        class E_type(SymEnum):
                            # sorted alphabetically
                            bidirectional = "bidirectional"
                            free = "free"
                            input = "input"
                            no_connect = "no_connect"
                            open_collector = "open_collector"
                            open_emitter = "open_emitter"
                            output = "output"
                            passive = "passive"
                            power_in = "power_in"
                            power_out = "power_out"
                            tri_state = "tri_state"
                            unspecified = "unspecified"

                        class E_style(SymEnum):
                            # sorted alphabetically
                            clock = "clock"
                            clock_low = "clock_low"
                            edge_clock_high = "edge_clock_high"
                            input_low = "input_low"
                            inverted = "inverted"
                            inverted_clock = "inverted_clock"
                            line = "line"
                            non_logic = "non_logic"
                            output_low = "output_low"

                        @dataclass
                        class C_name:
                            name: str = field(**sexp_field(positional=True))
                            effects: C_effects = field(default_factory=C_effects)

                        @dataclass
                        class C_number:
                            number: str = field(**sexp_field(positional=True))
                            effects: C_effects = field(default_factory=C_effects)

                        at: C_xyr
                        length: float
                        type: E_type = field(**sexp_field(positional=True))
                        style: E_style = field(**sexp_field(positional=True))
                        name: C_name = field(default_factory=C_name)
                        number: C_number = field(default_factory=C_number)

                    name: str = field(**sexp_field(positional=True))
                    polylines: list[C_polyline] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )
                    circles: list[C_circle] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )
                    rectangles: list[C_rect] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )
                    arcs: list[C_arc] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )
                    pins: list[C_pin] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )

                class E_hide(SymEnum):
                    hide = "hide"

                @dataclass
                class C_power:
                    pass

                name: str = field(**sexp_field(positional=True))
                power: Optional[C_power] = None
                propertys: dict[str, C_property] = field(
                    **sexp_field(multidict=True, key=lambda x: x.name),
                    default_factory=dict,
                )
                pin_numbers: Optional[E_hide] = None
                pin_names: Optional[C_pin_names] = None
                in_bom: Optional[bool] = None
                on_board: Optional[bool] = None
                symbols: dict[str, C_symbol] = field(
                    **sexp_field(multidict=True, key=lambda x: x.name),
                    default_factory=dict,
                )
                convert: Optional[int] = None

            symbols: dict[str, C_symbol] = field(
                **sexp_field(multidict=True, key=lambda x: x.name), default_factory=dict
            )

        @dataclass
        class C_symbol_instance:
            """
            TODO: Confusingly name, this isn't the class of
            "symbol_instances" at the top-level of the .kicad_sch file
            """

            @dataclass
            class C_pin:
                name: str = field(**sexp_field(positional=True))
                uuid: UUID = uuid_field()

            lib_id: str
            at: C_xyr
            unit: int
            in_bom: bool = False
            on_board: bool = False
            uuid: UUID = uuid_field()
            fields_autoplaced: bool = True
            propertys: dict[str, C_property] = field(
                **sexp_field(multidict=True, key=lambda x: x.name),
                default_factory=dict,
            )
            pins: list[C_pin] = field(
                **sexp_field(multidict=True), default_factory=list
            )
            convert: Optional[int] = None

        @dataclass
        class C_junction:
            at: C_xy
            diameter: float
            color: tuple[int, int, int, int]
            uuid: UUID = uuid_field()

        @dataclass
        class C_wire:
            pts: C_pts
            stroke: C_stroke
            uuid: UUID = uuid_field()

        @dataclass
        class C_text:
            at: C_xyr
            effects: C_effects
            text: str = field(**sexp_field(positional=True))
            uuid: UUID = uuid_field()

        @dataclass
        class C_sheet:
            @dataclass
            class C_pin:
                class E_type(SymEnum):
                    # sorted alphabetically
                    bidirectional = "bidirectional"
                    input = "input"
                    output = "output"
                    passive = "passive"
                    tri_state = "tri_state"

                at: C_xyr
                effects: C_effects
                name: str = field(**sexp_field(positional=True))
                type: E_type = field(**sexp_field(positional=True))
                uuid: UUID = uuid_field()

            at: C_xy
            size: C_xy
            stroke: C_stroke
            fill: C_fill
            uuid: UUID = uuid_field()
            fields_autoplaced: bool = True
            propertys: dict[str, C_property] = field(
                **sexp_field(multidict=True, key=lambda x: x.name),
                default_factory=dict,
            )
            pins: list[C_pin] = field(
                **sexp_field(multidict=True), default_factory=list
            )

        @dataclass
        class C_global_label:
            class E_shape(SymEnum):
                # sorted alphabetically
                input = "input"
                output = "output"
                bidirectional = "bidirectional"
                tri_state = "tri_state"
                passive = "passive"
                dot = "dot"
                round = "round"
                diamond = "diamond"
                rectangle = "rectangle"

            shape: E_shape
            at: C_xyr
            effects: C_effects
            text: str = field(**sexp_field(positional=True))
            uuid: UUID = uuid_field()
            fields_autoplaced: bool = True
            propertys: dict[str, C_property] = field(
                **sexp_field(multidict=True, key=lambda x: x.name),
                default_factory=dict,
            )

        # TODO: inheritance
        # text
        # label
        # global_label
        # hierarchical_label
        # netclass_flag
        # directive_label
        @dataclass
        class C_label:
            at: C_xyr
            effects: C_effects
            text: str = field(**sexp_field(positional=True))
            uuid: UUID = uuid_field()

        @dataclass
        class C_bus:
            pts: C_pts
            stroke: C_stroke
            uuid: UUID = uuid_field()

        @dataclass
        class C_bus_entry:
            at: C_xy
            size: C_xy
            stroke: C_stroke
            uuid: UUID = uuid_field()

        version: int = field(**sexp_field(assert_value=20211123))
        generator: str
        paper: str
        uuid: UUID = uuid_field()
        lib_symbols: C_lib_symbols = field(default_factory=C_lib_symbols)
        title_block: C_title_block = field(default_factory=C_title_block)

        junctions: list[C_junction] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        wires: list[C_wire] = field(**sexp_field(multidict=True), default_factory=list)

        texts: list[C_text] = field(**sexp_field(multidict=True), default_factory=list)
        symbols: list[C_symbol_instance] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        sheets: list[C_sheet] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        global_labels: list[C_global_label] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        no_connects: list[C_xy] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        buss: list[C_bus] = field(**sexp_field(multidict=True), default_factory=list)
        labels: list[C_label] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        bus_entrys: list[C_bus_entry] = field(
            **sexp_field(multidict=True), default_factory=list
        )

        # TODO: "symbol_instances" may or may not be required?
        # It appears to be a list of all symbols used in the schematic
        # But it also appears to be a translation of "symbols"

    kicad_sch: C_kicad_sch


@dataclass
class C_kicad_sym_file(SEXP_File):
    """
    When in doubt check: kicad/eeschema/sch_io/kicad_sexpr/sch_io_kicad_sexpr_parser.cpp
    """

    @dataclass
    class C_kicad_symbol_lib:
        version: int
        generator: str
        symbols: dict[str, C_kicad_sch_file.C_kicad_sch.C_lib_symbols.C_symbol] = field(
            **sexp_field(multidict=True, key=lambda x: x.name), default_factory=dict
        )

    kicad_symbol_lib: C_kicad_symbol_lib



import itertools
import logging
import re
import sys
import textwrap
import webbrowser
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import StrEnum, auto
from pathlib import Path
from typing import Annotated, Any, Callable, Iterator, cast

import caseconverter
import git
import questionary
import rich
import typer
from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter
from natsort import natsorted
from rich.table import Table

from atopile import errors, version
from atopile.address import AddrStr
from atopile.config import PROJECT_CONFIG_FILENAME, config
from atopile.telemetry import log_to_posthog
from faebryk.library.has_designator_prefix import has_designator_prefix
from faebryk.libs.picker.api.api import ApiHTTPError, Component
from faebryk.libs.picker.api.picker_lib import _extract_numeric_id
from faebryk.libs.picker.lcsc import download_easyeda_info
from faebryk.libs.pycodegen import (
    fix_indent,
    format_and_write,
    gen_block,
    gen_repeated_block,
    sanitize_name,
)
from faebryk.libs.util import groupby

# Set up logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


PROJECT_TEMPLATE = "https://github.com/atopile/project-template"

create_app = typer.Typer(rich_markup_mode="rich")


def help(text: str) -> None:  # pylint: disable=redefined-builtin
    """Print help text."""
    rich.print("\n" + textwrap.dedent(text).strip() + "\n")


def _stuck_user_helper() -> Iterator[bool]:
    """Figure out if a user is stuck and help them exit."""
    threshold = 5
    for i in itertools.count():
        if i >= threshold:
            if questionary.confirm("Are you trying to exit?").unsafe_ask():
                rich.print("No worries! Try Ctrl+C next time!")
                exit(0)
            threshold += 5
        yield True


stuck_user_helper_generator = _stuck_user_helper()


def _in_git_repo(path: Path) -> bool:
    """Check if the current directory is in a git repo."""
    try:
        git.Repo(path)
    except git.InvalidGitRepositoryError:
        return False
    return True


def query_helper[T: str | Path | bool](
    prompt: str,
    type_: type[T],
    clarifier: Callable[[Any], T] = lambda x: x,
    upgrader: Callable[[T], T] = lambda x: x,
    upgrader_msg: str | None = None,
    validator: str | Callable[[T], bool] | None = None,
    validation_failure_msg: str | None = "Value [cyan]{value}[/] is invalid",
    default: T | None = None,
    pre_entered: T | None = None,
    validate_default: bool = True,
) -> T:
    """Query a user for input."""
    rich.print(prompt)

    # Check the default value
    if default is not None:
        if not isinstance(default, type_):
            raise ValueError(f"Default value {default} is not of type {type_}")

    # Make a querier
    if type_ is str:

        def querier() -> str:
            return questionary.text(
                "",
                default=str(default or ""),
            ).unsafe_ask()

    elif type_ is Path:

        def querier() -> Path:
            return Path(
                questionary.path(
                    "",
                    default=str(default or ""),
                ).unsafe_ask()
            )

    elif type_ is bool:
        assert default is None or isinstance(default, bool)

        def querier() -> bool:
            return questionary.confirm(
                "",
                default=default or True,
            ).unsafe_ask()

    else:
        raise ValueError(f"Unsupported query type: `{type_}`")

    # Ensure a validator
    # Default the validator to a regex match if it's a str
    if validator is None:

        def validator_func(value: T) -> bool:
            return True

    elif isinstance(validator, str):

        def validator_func(value: T) -> bool:
            return bool(re.match(validator, value))  # type: ignore

        if not validation_failure_msg:
            validation_failure_msg = f'Value must match regex: `r"{validator}"`'

    else:
        validator_func = validator  # type: ignore

    # Ensure the default provided is valid
    if default is not None and validate_default:
        if not validator_func(clarifier(default)):
            logger.debug(f"Default value {default} is invalid")

        if clarifier(default) != upgrader(clarifier(default)):
            logger.debug(f"Default value {default} doesn't meet best-practice")

        default = upgrader(clarifier(default))

    # When running non-interactively, we expect the value to be provided
    # at the command level, so we don't need to query the user for it
    # Validate and return the pre-entered value
    if not config.interactive:
        if pre_entered is None:
            raise ValueError("Value is required. Check at command level.")

        if not validator_func(pre_entered):
            if validation_failure_msg:
                msg = validation_failure_msg.format(value=pre_entered)
            else:
                msg = f"Value {pre_entered} is invalid"
            raise errors.UserException(msg)

        return pre_entered

    # Pre-entered values are expected to skip the query for a value in the first place
    # but progress through the validator and upgrader if we're running interactively
    value: T | None = pre_entered

    for _ in stuck_user_helper_generator:
        if value is None:
            value = clarifier(querier())  # type: ignore
        assert isinstance(value, type_)

        if (proposed_value := upgrader(value)) != value:
            if upgrader_msg:
                rich.print(upgrader_msg.format(proposed_value=proposed_value))

            rich.print(f"Use [cyan]{proposed_value}[/] instead?")
            if questionary.confirm("").unsafe_ask():
                value = proposed_value

        if not validator_func(value):
            if validation_failure_msg:
                rich.print(validation_failure_msg.format(value=value))
            value = None
            continue

        return value

    raise RuntimeError("Unclear how we got here")


PROJECT_NAME_REQUIREMENTS = (
    "Project name must start with a letter and contain only letters, numbers, dashes"
    " and underscores. It will be used for the project directory and name on GitHub"
)


@create_app.command()
@log_to_posthog("cli:create_project_end")
def project(
    template: str = "https://github.com/atopile/project-template @ compiler-v0.4",
    create_github_repo: bool | None = None,
):
    """
    Create a new ato project.
    """
    # TODO: add template options

    template_ref, *template_branch = template.split("@")
    template_ref = template_ref.strip()
    if template_branch:
        template_branch = template_branch[0].strip()
    else:
        template_branch = None

    if create_github_repo is True and not config.interactive:
        raise errors.UserException(
            "Cannot create a GitHub repo when running non-interactively."
        )

    extra_context = {
        "__ato_version": version.get_installed_atopile_version(),
        "__python_path": sys.executable,
    }

    logging.info("Running cookie-cutter on the template")
    try:
        project_path = Path(
            cookiecutter(
                template_ref,
                checkout=template_branch,
                no_input=not config.interactive,
                extra_context=dict(
                    filter(lambda x: x[1] is not None, extra_context.items())
                ),
            )
        )
    except OutputDirExistsException as e:
        raise errors.UserException(
            "Directory already exists. Please choose a different name."
        ) from e

    if create_github_repo is None:
        create_github_repo = (
            query_helper(
                "Create a GitHub repo for this project?",
                bool,
                default=True,
            )
            if config.interactive
            else False
        )

    # Get a repo
    if create_github_repo:
        logging.info("Initializing git repo")
        repo = git.Repo.init(project_path)
        repo.git.add(A=True, f=True)
        repo.git.commit(m="Initial commit")

        github_username = query_helper(
            "What's your GitHub username?",
            str,
            validator=r"^[a-zA-Z0-9_-]+$",
        )

        make_repo_url = (
            f"https://github.com/new?name={project_path.name}&owner={github_username}"
        )

        help(
            f"""
            We recommend you create a GitHub repo for your project.

            If you already have a repo, you can respond [yellow]n[/]
            to the next question and provide the URL to your repo.

            If you don't have one, you can respond yes to the next question
            or (Cmd/Ctrl +) click the link below to create one.

            Just select the template you want to use.

            {make_repo_url}
            """
        )

        webbrowser.open(make_repo_url)

        def _repo_validator(url: str) -> bool:
            try:
                repo.create_remote("origin", url).push()
                return True
            except Exception:
                return False

        query_helper(
            ":rocket: What's the [cyan]repo's URL?[/]",
            str,
            default=f"https://github.com/{github_username}/{project_path.name}",
            validator=_repo_validator,
            validation_failure_msg="Remote could not be added: {value}",
            validate_default=False,
        )

    # Wew! New repo created!
    rich.print(
        f':sparkles: [green]Created new project "{project_path.name}"![/] :sparkles:'
    )


@create_app.command("build-target")
@log_to_posthog("cli:create_build_target_end")
def build_target(
    build_target: Annotated[str | None, typer.Option()] = None,
    file: Annotated[Path | None, typer.Option()] = None,
    module: Annotated[str | None, typer.Option()] = None,
    backup_name: Annotated[str | None, typer.Argument()] = None,
):
    """
    Create a new build configuration.
    - adds entry to ato.yaml
    - creates a new directory in layout
    """
    config.apply_options(None)

    try:
        src_path = config.project.paths.src
        config.project_dir  # touch property to ensure config's loaded from a project
    except ValueError:
        raise errors.UserNoProjectException()

    if backup_name is None:
        if build_target:
            backup_name = build_target
        elif file:
            backup_name = file.name.split(".", 1)[0]
        elif module:
            backup_name = module

    # If we're running non-interactively, all details must be provided
    if not config.interactive and not all([build_target, file, module]):
        raise errors.UserException(
            "--build-target, --file, and --module must all be"
            " provided when running non-interactively."
        )

    def _check_build_target_name(value: str) -> bool:
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9_-]*$", value):
            rich.print(
                "[red]Build-target names must start with a letter and"
                " contain only letters, numbers, dashes and underscores.[/]"
            )
            return False

        if value in config.project.builds:
            rich.print(f"[red]Build-target `{value}` already exists[/]")
            return False

        return True

    build_target = query_helper(
        ":rocket: What's the [cyan]build-target[/] name?",
        type_=str,
        upgrader=caseconverter.kebabcase,
        upgrader_msg="We recommend using kebab-case for build-target names.",
        validator=_check_build_target_name,
        validation_failure_msg="",
        pre_entered=build_target,
        default=caseconverter.kebabcase(backup_name) if backup_name else None,
    )

    def _file_clarifier(value: Path) -> Path:
        if not value.is_absolute():
            # If it's a relative path wrt ./ (cwd) then respect that
            # else assume it's relative to the src directory
            if str(value).startswith("./") or str(value).startswith(".\\"):
                value = Path(value).resolve()
            else:
                value = src_path / value

        return value

    def _file_updator(value: Path) -> Path:
        value = value.with_stem(caseconverter.snakecase(value.stem))
        if value.suffix != ".ato":
            # Allow dots in filenames
            value = value.with_suffix(value.suffix + ".ato")
        return value

    def _file_validator(f: Path) -> bool:
        if f.is_dir():
            rich.print(f"{f} is a directory")
            return False

        if f.suffix != ".ato":
            rich.print(f"{f} must end in .ato")
            return False

        try:
            f.relative_to(src_path)
        except ValueError:
            rich.print(f"{f} is outside the project's src dir")
            return False

        return True

    file = query_helper(
        ":rocket: What [cyan]file[/] should we add the module to?",
        type_=Path,
        clarifier=_file_clarifier,
        upgrader=_file_updator,
        upgrader_msg=(
            "We recommend using snake_case for file names, and it must end in .ato"
        ),
        validator=_file_validator,
        validation_failure_msg="",
        pre_entered=file,
        default=Path(caseconverter.snakecase(backup_name or build_target) + ".ato"),
    )

    module = query_helper(
        ":rocket: What [cyan]module[/] should we add to the file?",
        type_=str,
        validator=str.isidentifier,
        validation_failure_msg="",
        upgrader=caseconverter.pascalcase,
        upgrader_msg="We recommend using pascal-case for module names.",
        pre_entered=module,
        default=caseconverter.pascalcase(backup_name or build_target),
    )

    logger.debug(f"Creating build-target with {build_target=}, {file=}, {module=}")

    # Update project config
    logger.info(
        f"Adding build-target to {config.project.paths.root / PROJECT_CONFIG_FILENAME}"
    )

    def add_build_target(config_data: dict, new_data: dict):
        config_data["builds"][build_target] = new_data
        return config_data

    config.update_project_settings(
        add_build_target, {"entry": str(AddrStr.from_parts(file, module))}
    )

    # Create or add to file
    module_text = f"module {module}:\n    pass\n"

    if file.is_file():  # exists and is a file
        with file.open("a", encoding="utf-8") as f:
            f.write("\n")
            f.write(module_text)

    else:
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(module_text, encoding="utf-8")

    rich.print(
        ":sparkles: Successfully created a new build configuration "
        f"[cyan]{build_target}[/] at [cyan]{file}[/]! :sparkles:"
    )


class ComponentType(StrEnum):
    ato = auto()
    fab = auto()


@create_app.command()
@log_to_posthog("cli:create_component_end")
def component(
    search_term: Annotated[str | None, typer.Option("--search", "-s")] = None,
    name: Annotated[str | None, typer.Option("--name", "-n")] = None,
    filename: Annotated[str | None, typer.Option("--filename", "-f")] = None,
    type_: Annotated[ComponentType | None, typer.Option("--type", "-t")] = None,
):
    """Create a new component."""
    from faebryk.libs.picker.api.models import Component
    from faebryk.libs.picker.api.picker_lib import client
    from faebryk.libs.pycodegen import sanitize_name

    config.apply_options(None)

    # Find a component --------------------------------------------------------

    component: Component | None = None

    for _ in stuck_user_helper_generator:
        if not search_term:
            search_term = questionary.text(
                "Search for a component (Part Number or LCSC ID):"
            ).unsafe_ask()
            assert search_term is not None

        try:
            lcsc_id = _extract_numeric_id(search_term)
        except ValueError:
            lcsc_id = None

        try:
            if lcsc_id:
                components = client.fetch_part_by_lcsc(lcsc_id)
            else:
                # TODO: remove this once we have a fuzzy search
                mfr = questionary.text("Enter the manufacturer").unsafe_ask()
                components = client.fetch_part_by_mfr(mfr, search_term)
        except ApiHTTPError as e:
            if e.response.status_code == 404:
                components = []
            else:
                raise

        if len(components) == 0:
            rich.print(f'No components found for "{search_term}"')
            search_term = None
            continue

        component_table = Table()
        component_table.add_column("Part Number")
        component_table.add_column("Manufacturer")
        component_table.add_column("Description")

        for component in components:
            component_table.add_row(
                component.manufacturer_name,
                component.part_number,
                component.description,
            )

        rich.print(component_table)

        choices = [
            {
                "name": f"{component.manufacturer_name} {component.part_number}",
                "value": component,
            }
            for component in components
        ] + [{"name": "Search again...", "value": None}]

        component = questionary.select(
            "Select a component", choices=choices
        ).unsafe_ask()

        if component is not None:
            break

        # Reset the input terms to start over if we didn't find what we're looking for
        search_term = None

    # We have a component -----------------------------------------------------
    assert component is not None

    # TODO: templated ato components too
    # if type_ is None:
    #     type_ = ComponentType.fab
    if type_ is None:
        type_ = questionary.select(
            "Select the component type", choices=list(ComponentType)
        ).unsafe_ask()
        assert type_ is not None

    if name is None:
        name = questionary.text(
            "Enter the name of the component",
            default=caseconverter.pascalcase(
                sanitize_name(component.manufacturer_name + " " + component.part_number)
            ),
        ).unsafe_ask()

    sanitized_name = sanitize_name(name)
    if sanitized_name != name:
        rich.print(f"Sanitized name: {sanitized_name}")

    if type_ == ComponentType.ato:
        extension = ".ato"
    elif type_ == ComponentType.fab:
        extension = ".py"
    else:
        raise ValueError(f"Invalid component type: {type_}")

    out_path: Path | None = None
    for _ in stuck_user_helper_generator:
        if filename is None:
            filename = questionary.text(
                "Enter the filename of the component",
                default=caseconverter.snakecase(name) + extension,
            ).unsafe_ask()

        assert filename is not None

        filepath = Path(filename)
        if filepath.is_absolute():
            out_path = filepath.resolve()
        else:
            out_path = (config.project.paths.src / filename).resolve()

        if out_path.exists():
            rich.print(f"File {out_path} already exists")
            filename = None
            continue

        if not out_path.parent.exists():
            rich.print(
                f"Directory {out_path.parent} does not exist. Creating it now..."
            )
            out_path.parent.mkdir(parents=True, exist_ok=True)

        break

    assert out_path is not None

    if type_ == ComponentType.ato:
        template = AtoTemplate(name=sanitized_name, base="Module")
        template.add_part(component)
        out = template.dumps()
        out_path.write_text(out, encoding="utf-8")
        rich.print(f":sparkles: Created {out_path} !")

    elif type_ == ComponentType.fab:
        template = FabllTemplate(name=sanitized_name, base="Module")
        template.add_part(component)
        out = template.dumps()
        format_and_write(out, out_path)
        rich.print(f":sparkles: Created {out_path} !")


@create_app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    if ctx.resilient_parsing:
        return

    if not ctx.invoked_subcommand:
        commands = cast(dict, ctx.command.commands)  # type: ignore  # commands is an attribute of the context
        command_name = questionary.select(
            "What would you like to create?",
            choices=[n for n, c in commands.items() if not c.hidden],
        ).unsafe_ask()

        assert command_name in commands

        # Run the command
        ctx.invoke(commands[command_name].callback)


if __name__ == "__main__":
    create_app()  # pylint: disable=no-value-for-parameter


@dataclass
class CTX:
    path: Path
    pypath: str
    overwrite: bool


@dataclass
class Template(ABC):
    name: str
    base: str
    imports: list[str] = field(default_factory=list)
    nodes: list[str] = field(default_factory=list)
    docstring: str = "TODO: Docstring describing your module"

    def _process_part(self, part: Component):
        """Common part processing logic used by child classes."""
        name = sanitize_name(f"{part.manufacturer_name}_{part.part_number}")
        assert isinstance(name, str)
        _, _, _, _, easyeda_symbol, _ = download_easyeda_info(
            part.lcsc_display, get_model=False
        )
        return name, easyeda_symbol

    @abstractmethod
    def add_part(self, part: Component): ...


@dataclass
class AtoTemplate(Template):
    attributes: list[str] = field(default_factory=list)
    pins: list[str] = field(default_factory=list)
    defined_signals: set[str] = field(default_factory=set)

    def add_part(self, part: Component):
        # Get common processed data
        self.name, easyeda_symbol = self._process_part(part)

        # Set docstring with description
        self.docstring = part.description

        # Add component metadata
        self.attributes.extend(
            [
                f'lcsc_id = "{part.lcsc_display}"',
                f'manufacturer = "{part.manufacturer_name}"',
                f'mpn = "{part.part_number}"',
            ]
        )

        # Add datasheet if available
        if part.datasheet_url:
            self.attributes.append(f'datasheet_url = "{part.datasheet_url}"')

        # Add designator prefix from EasyEDA symbol
        designator_prefix = easyeda_symbol.info.prefix.replace("?", "")
        self.attributes.append(f'designator_prefix = "{designator_prefix}"')

        # Collect and sort pins first
        sorted_pins = []
        if hasattr(easyeda_symbol, "units") and easyeda_symbol.units:
            for unit in easyeda_symbol.units:
                if hasattr(unit, "pins"):
                    for pin in unit.pins:
                        pin_num = pin.settings.spice_pin_number
                        pin_name = pin.name.text
                        if (
                            pin_name
                            and pin_name not in ["NC", "nc"]
                            and not re.match(r"^[0-9]+$", pin_name)
                        ):
                            sorted_pins.append((sanitize_name(pin_name), pin_num))

        # Sort pins by name using natsort
        sorted_pins = natsorted(sorted_pins, key=lambda x: x[0])

        # Process sorted pins
        for pin_name, pin_num in sorted_pins:
            if pin_name not in self.defined_signals:
                self.pins.append(f"signal {pin_name} ~ pin {pin_num}")
                self.defined_signals.add(pin_name)
            else:
                self.pins.append(f"{pin_name} ~ pin {pin_num}")

    def dumps(self) -> str:
        output = f"component {self.name}:\n"
        output += f'    """{self.name} component"""\n'

        # Add attributes
        for attr in self.attributes:
            output += f"    {attr}\n"

        # Add blank line after attributes
        output += "\n"

        if self.pins:
            output += "    # pins\n"
            for pin in self.pins:
                output += f"    {pin}\n"

        return output


@dataclass
class FabllTemplate(Template):
    traits: list[str] = field(default_factory=list)

    def add_part(self, part: Component):
        # Get common processed data
        self.name, easyeda_symbol = self._process_part(part)

        designator_prefix_str = easyeda_symbol.info.prefix.replace("?", "")
        try:
            prefix = has_designator_prefix.Prefix(designator_prefix_str)
            designator_prefix = f"F.has_designator_prefix.Prefix.{prefix.name}"
        except ValueError:
            logger.warning(
                f"Using non-standard designator prefix: {designator_prefix_str}"
            )
            designator_prefix = f"'{designator_prefix_str}'"

        self.traits.append(
            f"designator_prefix = L.f_field(F.has_designator_prefix)"
            f"({designator_prefix})"
        )

        self.traits.append(
            "lcsc_id = L.f_field(F.has_descriptive_properties_defined)"
            f"({{'LCSC': '{part.lcsc_display}'}})"
        )

        self.imports.append(
            "from faebryk.libs.picker.picker import DescriptiveProperties"
        )
        self.traits.append(
            f"descriptive_properties = L.f_field(F.has_descriptive_properties_defined)"
            f"({{DescriptiveProperties.manufacturer: '{part.manufacturer_name}', "
            f"DescriptiveProperties.partno: '{part.part_number}'}})"
        )

        if url := part.datasheet_url:
            self.traits.append(
                f"datasheet = L.f_field(F.has_datasheet_defined)('{url}')"
            )

        partdoc = part.description.replace("  ", "\n")
        self.docstring = f"{self.docstring}\n\n{partdoc}"

        # pins --------------------------------
        no_name: list[str] = []
        no_connection: list[str] = []
        interface_names_by_pin_num: dict[str, str] = {}

        for unit in easyeda_symbol.units:
            for pin in unit.pins:
                pin_num = pin.settings.spice_pin_number
                pin_name = pin.name.text
                if re.match(r"^[0-9]+$", pin_name):
                    no_name.append(pin_num)
                elif pin_name in ["NC", "nc"]:
                    no_connection.append(pin_num)
                else:
                    pyname = sanitize_name(pin_name)
                    interface_names_by_pin_num[pin_num] = pyname

        self.nodes.append(
            "#TODO: Change auto-generated interface types to actual high level types"
        )

        _interface_lines_by_min_pin_num = {}
        for interface_name, _items in groupby(
            interface_names_by_pin_num.items(), lambda x: x[1]
        ).items():
            pin_nums = [x[0] for x in _items]
            line = f"{interface_name}: F.Electrical  # {'pin' if len(pin_nums) == 1 else 'pins'}: {', '.join(pin_nums)}"  # noqa: E501
            _interface_lines_by_min_pin_num[min(pin_nums)] = line
        self.nodes.extend(
            line
            for _, line in natsorted(
                _interface_lines_by_min_pin_num.items(), key=lambda x: x[0]
            )
        )

        if no_name:
            self.nodes.append(f"unnamed = L.list_field({len(no_name)}, F.Electrical)")

        pin_lines = (
            [
                f'"{pin_num}": self.{interface_name},'
                for pin_num, interface_name in interface_names_by_pin_num.items()
            ]
            + [f'"{pin_num}": None,' for pin_num in no_connection]
            + [f'"{pin_num}": self.unnamed[{i}],' for i, pin_num in enumerate(no_name)]
        )
        self.traits.append(
            fix_indent(f"""
            @L.rt_field
            def attach_via_pinmap(self):
                return F.can_attach_to_footprint_via_pinmap(
                    {{
                        {gen_repeated_block(natsorted(pin_lines))}
                    }}
                )
        """)
        )

    def dumps(self) -> str:
        always_import = [
            "import faebryk.library._F as F  # noqa: F401",
            f"from faebryk.core.{self.base.lower()} import {self.base}",
            "from faebryk.libs.library import L  # noqa: F401",
            "from faebryk.libs.units import P  # noqa: F401",
        ]

        self.imports = always_import + self.imports

        out = fix_indent(f"""
            # This file is part of the faebryk project
            # SPDX-License-Identifier: MIT

            import logging

            {gen_repeated_block(self.imports)}

            logger = logging.getLogger(__name__)

            class {self.name}({self.base}):
                \"\"\"
                {gen_block(self.docstring)}
                \"\"\"

                # ----------------------------------------
                #                modules
                # ----------------------------------------

                # ----------------------------------------
                #              interfaces
                # ----------------------------------------
                {gen_repeated_block(self.nodes)}

                # ----------------------------------------
                #              parameters
                # ----------------------------------------

                # ----------------------------------------
                #                traits
                # ----------------------------------------
                {gen_repeated_block(self.traits)}

                def __preinit__(self):
                    # ------------------------------------
                    #           connections
                    # ------------------------------------

                    # ------------------------------------
                    #          parametrization
                    # ------------------------------------
                    pass
        """)

        return out



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class SK6812(Module):
    """
    RGB digital LED with SK6812 controller

    SMD5050-4P
    RGB LEDs(Built-in IC) ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    power: F.ElectricPower
    data_in: F.ElectricLogic
    data_out: F.ElectricLogic

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.data_in, self.data_out)

    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C5378720"})
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.LED
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "OPSCO Optoelectronics",
            DescriptiveProperties.partno: "SK6812",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/2303300930_OPSCO-Optoelectronics-SK6812_C5378720.pdf"  # noqa: E501
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.data_in.line: ["DIN"],
                self.data_out.line: ["DOUT"],
                self.power.lv: ["VSS", "GND"],
                self.power.hv: ["VDD"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        # FIXME
        # self.power.decoupled.decouple()
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_module_references(self, exclude=[self.power])

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power.voltage.constrain_subset(L.Range(3.3 * P.V, 5.5 * P.V))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
from pathlib import Path

from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class has_reference_layout(L.Module.TraitT.decless()):
    """
    The module has a reference layout.
    """

    def __init__(self, *paths: os.PathLike):
        self._paths = set(Path(p).expanduser().resolve().absolute() for p in paths)

    @property
    def paths(self) -> set[Path]:
        return self._paths

    def handle_duplicate(self, old: "has_reference_layout", _: L.Node) -> bool:
        old.paths.update(self.paths)
        return False



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import Sequence
from warnings import deprecated

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L


@deprecated("Use F.has_package instead")
class has_package_requirement(Module.TraitT.decless()):
    def __init__(self, *package_candidates: str) -> None:
        super().__init__()
        self._package_candidates = list(package_candidates)

    def on_obj_set(self):
        obj = self.get_obj(L.Module)
        obj.add(F.has_package(*self._package_candidates))
        return super().on_obj_set()

    # Delete this
    def get_package_candidates(self) -> Sequence[str]:
        raise NotImplementedError("Use F.has_package instead")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import unittest

import faebryk.library._F as F
from faebryk.libs.units import P


class TestFootprints(unittest.TestCase):
    def test_qfn_kicad(self):
        test_cases = {
            "Package_DFN_QFN:QFN-16-1EP_4x4mm_P0.5mm_EP2.45x2.45mm_ThermalVias": F.QFN(  # noqa: E501
                pin_cnt=16,
                exposed_thermal_pad_cnt=1,
                size_xy=(4 * P.mm, 4 * P.mm),
                pitch=0.5 * P.mm,
                exposed_thermal_pad_dimensions=(2.45 * P.mm, 2.45 * P.mm),
                has_thermal_vias=True,
            ),
            "Package_DFN_QFN:QFN-12-1EP_3x3mm_P0.5mm_EP1.6x1.6mm": F.QFN(
                pin_cnt=12,
                exposed_thermal_pad_cnt=1,
                size_xy=(3 * P.mm, 3 * P.mm),
                pitch=0.5 * P.mm,
                exposed_thermal_pad_dimensions=(1.6 * P.mm, 1.6 * P.mm),
                has_thermal_vias=False,
            ),
            "Package_DFN_QFN:QFN-20-1EP_3.5x3.5mm_P0.5mm_EP2x2mm": F.QFN(
                pin_cnt=20,
                exposed_thermal_pad_cnt=1,
                size_xy=(3.5 * P.mm, 3.5 * P.mm),
                pitch=0.5 * P.mm,
                exposed_thermal_pad_dimensions=(2 * P.mm, 2 * P.mm),
                has_thermal_vias=False,
            ),
        }

        for solution, footprint in test_cases.items():
            self.assertEqual(
                solution,
                footprint.get_trait(F.has_kicad_footprint).get_kicad_footprint(),
            )

    def test_qfn_constraints(self):
        # test good done in qfn kicad
        # test bad

        # only thermal via if thermal pad cnt > 0
        self.assertRaises(
            AssertionError,
            lambda: F.QFN(
                pin_cnt=20,
                exposed_thermal_pad_cnt=0,
                size_xy=(3.5 * P.mm, 3.5 * P.mm),
                pitch=0.5 * P.mm,
                exposed_thermal_pad_dimensions=(2.5 * P.mm, 2.5 * P.mm),
                has_thermal_vias=True,
            ),
        )

        # pad has to be smaller than package
        self.assertRaises(
            AssertionError,
            lambda: F.QFN(
                pin_cnt=20,
                exposed_thermal_pad_cnt=1,
                size_xy=(3.5 * P.mm, 3.5 * P.mm),
                pitch=0.5 * P.mm,
                exposed_thermal_pad_dimensions=(4.5 * P.mm, 2.5 * P.mm),
                has_thermal_vias=False,
            ),
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class App(Module):
    lowpass: F.Filter

    def __preinit__(self) -> None:
        # TODO actually do something with the filter

        # Specialize
        special = self.lowpass.specialize(F.FilterElectricalRC())

        # Parametrize
        special.cutoff_frequency.constrain_subset(
            L.Range.from_center_rel(1 * P.kHz, 0.1)
        )
        special.response.constrain_subset(F.Filter.Response.LOWPASS)
        special.resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1.05 * P.kohm, 0.01)
        )

        # set reference voltage
        # TODO: this will be automatically set by the power supply
        # once this example is more complete
        special.in_.reference.voltage.constrain_subset(
            L.Range.from_center_rel(3 * P.V, 0.05)
        )
        special.out.reference.voltage.constrain_subset(
            L.Range.from_center_rel(3 * P.V, 0.05)
        )

        # TODO
        # Construct
        # special.get_trait(F.has_construction_dependency).construct()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import unittest

from faebryk.core.cpp import (
    GraphInterface,
    GraphInterfaceHierarchical,
    LinkExists,
    LinkNamedParent,
)
from faebryk.core.link import (
    LinkDirect,
    LinkDirectConditional,
    LinkDirectConditionalFilterResult,
    LinkParent,
    LinkSibling,
)
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.libs.library import L


class TestGraph(unittest.TestCase):
    def test_gifs(self):
        from faebryk.core.graphinterface import GraphInterface as GIF

        gif1 = GIF()
        gif2 = GIF()

        gif1.connect(gif2)

        self.assertIn(gif2, gif1.edges)
        self.assertTrue(gif1.is_connected_to(gif2))

        gif3 = GIF()

        class linkcls(LinkDirect):
            pass

        gif1.connect(gif3, linkcls())
        self.assertIsInstance(gif1.is_connected_to(gif3), linkcls)
        self.assertEqual(gif1.is_connected_to(gif3), gif3.is_connected_to(gif1))

        self.assertRaises(LinkExists, lambda: gif1.connect(gif3))
        self.assertRaises(LinkExists, lambda: gif1.connect(gif3, linkcls()))

        self.assertEqual(gif1.G, gif2.G)

    def test_node_gifs(self):
        from faebryk.core.node import Node

        n1 = Node()

        self.assertIsInstance(n1.self_gif.is_connected_to(n1.parent), LinkSibling)
        self.assertIsInstance(n1.self_gif.is_connected_to(n1.children), LinkSibling)

        n2 = Node()
        n1.add(n2, name="n2")

        self.assertIsInstance(n1.children.is_connected_to(n2.parent), LinkParent)

        print(n1.get_graph())

        p = n2.get_parent()
        self.assertIsNotNone(p)
        assert p is not None
        self.assertIs(p[0], n1)

        self.assertEqual(n1.self_gif.G, n2.self_gif.G)

    # TODO move to own file
    def test_fab_ll_simple_hierarchy(self):
        class N(Node):
            SN1: Node
            SN2: Node
            SN3 = L.list_field(2, Node)

            @L.rt_field
            def SN4(self):
                return Node()

        n = N()
        children = n.get_children(direct_only=True, types=Node)
        self.assertEqual(children, {n.SN1, n.SN2, n.SN3[0], n.SN3[1], n.SN4})

    def test_fab_ll_chain_names(self):
        root = Node()
        x = root
        for i in range(10):
            y = Node()
            x.add(y, f"i{i}")
            x = y

        self.assertRegex(
            x.get_full_name(), "[*][0-9A-F]{4}.i0.i1.i2.i3.i4.i5.i6.i7.i8.i9"
        )

    def test_fab_ll_chain_tree(self):
        root = Node()
        x = root
        for i in range(10):
            y = Node()
            z = Node()
            x.add(y, f"i{i}")
            x.add(z, f"j{i}")
            x = y

        self.assertRegex(
            x.get_full_name(), "[*][0-9A-F]{4}.i0.i1.i2.i3.i4.i5.i6.i7.i8.i9"
        )

    def test_fab_ll_chain_tree_with_root(self):
        root = Node()
        root.no_include_parents_in_full_name = True
        x = root
        for i in range(10):
            y = Node()
            z = Node()
            x.add(y, f"i{i}")
            x.add(z, f"j{i}")
            x = y

        self.assertEqual(x.get_full_name(), "i0.i1.i2.i3.i4.i5.i6.i7.i8.i9")

    def test_link_eq_direct(self):
        gif1 = GraphInterface()
        gif2 = GraphInterface()

        gif1.connect(gif2)

        self.assertEqual(gif1.is_connected_to(gif2), LinkDirect())
        self.assertNotEqual(gif1.is_connected_to(gif2), LinkSibling())

    def test_link_eq_args(self):
        gif1 = GraphInterfaceHierarchical(is_parent=True)
        gif2 = GraphInterfaceHierarchical(is_parent=False)

        gif1.connect(gif2, link=LinkNamedParent("bla"))

        self.assertEqual(gif1.is_connected_to(gif2), LinkNamedParent("bla"))
        self.assertNotEqual(gif1.is_connected_to(gif2), LinkNamedParent("blub"))
        self.assertNotEqual(gif1.is_connected_to(gif2), LinkDirect())

    def test_inherited_link(self):
        class _Link(LinkDirectConditional):
            def __init__(self):
                super().__init__(
                    lambda path: LinkDirectConditionalFilterResult.FILTER_PASS,
                    needs_only_first_in_path=True,
                )

        gif1 = GraphInterface()
        gif2 = GraphInterface()

        gif1.connect(gif2, link=_Link())
        self.assertIsInstance(gif1.is_connected_to(gif2), _Link)

    def test_unique_mif_shallow_link(self):
        class MIFType(ModuleInterface):
            pass

        assert MIFType.LinkDirectShallow() is MIFType.LinkDirectShallow()

    def test_node_preinit(self):
        counter = 0

        def assert_and_reset(target: int):
            nonlocal counter
            self.assertEqual(counter, target)
            counter = 0

        class N1(Node):
            def __preinit__(self):
                nonlocal counter
                counter += 1

        class N11(N1):
            def __preinit__(self):
                nonlocal counter
                counter += 1

        class N12(N1):
            pass

        class N111(N11):
            def __preinit__(self):
                nonlocal counter
                counter += 1

        class N112(N11):
            pass

        N1()
        assert_and_reset(1)
        N11()
        assert_and_reset(2)
        N12()
        assert_and_reset(1)
        N111()
        assert_and_reset(3)
        N112()
        assert_and_reset(2)


def test_get_children_modules_simple():
    class App(Module):
        m: Module

    app = App()

    mods = app.get_children_modules(types=Module)
    assert mods == {app.m}


def test_get_children_modules_specialized():
    class App(Module):
        m: Module

    class ModuleSpecial(Module):
        pass

    app = App()
    special = ModuleSpecial()
    app.m.specialize(special)

    mods = app.get_children_modules(types=ModuleSpecial)
    assert mods == {special}


def test_get_children_modules_specialized_chain():
    class App(Module):
        m: Module

    class ModuleSpecial(Module):
        m: Module

    app = App()
    special = ModuleSpecial()
    app.m.specialize(special)

    special2 = ModuleSpecial()
    special.m.specialize(special2)

    mods = app.get_children_modules(types=ModuleSpecial)
    assert mods == {special, special2}

    mods_raw = app.get_children_modules(
        types=Module, f_filter=lambda x: type(x) is Module
    )
    assert mods_raw == {special2.m}


def test_get_children_modules_tree():
    import faebryk.library._F as F

    cap1 = F.Capacitor()
    cap2 = F.Capacitor()
    cap_1 = F.MultiCapacitor.from_capacitors(cap1, cap2)
    cap3 = F.MultiCapacitor(4)
    cap_2 = F.MultiCapacitor.from_capacitors(cap_1, cap3)

    mods = cap_1.get_children_modules(
        types=F.Capacitor, f_filter=lambda x: type(x) is F.Capacitor
    )
    assert mods == {cap1, cap2}

    mods = cap_2.get_children_modules(
        types=F.Capacitor, f_filter=lambda x: type(x) is F.Capacitor
    )
    assert mods == {cap1, cap2, *cap3.capacitors}



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Resistor(Module):
    unnamed = L.list_field(2, F.Electrical)

    resistance = L.p_field(units=P.ohm)
    max_power = L.p_field(units=P.W)
    max_voltage = L.p_field(units=P.V)

    attach_to_footprint: F.can_attach_to_footprint_symmetrically
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.R
    )

    @L.rt_field
    def pickable(self) -> F.is_pickable_by_type:
        return F.is_pickable_by_type(
            F.is_pickable_by_type.Type.Resistor,
            {
                "resistance": self.resistance,
                "max_power": self.max_power,
                "max_voltage": self.max_voltage,
            },
        )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(*self.unnamed)

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.resistance, tolerance=True),
            S(self.max_power),
        )

    def allow_removal_if_zero(self):
        # FIXME: enable as soon as solver works
        return
        # import faebryk.library._F as F

        # @once
        # def do_replace():
        #     self.resistance.constrain_subset(0.0 * P.ohm)
        #     self.unnamed[0].connect(self.unnamed[1])
        #     self.add(F.has_part_removed())
        #
        # self.resistance.operation_is_superset(0.0 * P.ohm).if_then_else(
        #     lambda: do_replace(),
        #     lambda: None,
        #     preference=True,
        # )
        #
        # def replace_zero(m: Module, solver: Solver):
        #     assert m is self
        #
        #     solver.assert_any_predicate(
        #         [(Is(self.resistance, 0.0 * P.ohm), None)], lock=True
        #     )
        #
        # self.add(
        #    F.has_multi_picker(-100, F.has_multi_picker.FunctionPicker(replace_zero))
        # )

    # TODO: remove @https://github.com/atopile/atopile/issues/727
    @property
    def p1(self) -> F.Electrical:
        """One side of the resistor."""
        return self.unnamed[0]

    @property
    def p2(self) -> F.Electrical:
        """The other side of the resistor."""
        return self.unnamed[1]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P


class ME6211C33M5G_N(F.LDO):
    """
    3.3V 600mA LDO
    """

    # components

    def __init__(self, default_enabled: bool = True) -> None:
        super().__init__()
        self._default_enabled = default_enabled

    def __preinit__(self):
        # set constraints
        self.output_voltage.constrain_superset(L.Range.from_center_rel(3.3 * P.V, 0.02))

        if self._default_enabled:
            self.enable.set(True)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power_in.hv,
                "2": self.power_in.lv,
                "3": self.enable.get_enable_signal(),
                "5": self.power_out.hv,
            }
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheet.lcsc.com/lcsc/1811131510_MICRONE-Nanjing-Micro-One-Elec-ME6211C33M5G-N_C82942.pdf"
    )

    mfr = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "MICRONE(Nanjing Micro One Elec)",
            DescriptiveProperties.partno: "ME6211C33M5G-N",
        }
    )

    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C82942"})



import contextlib
import logging
from abc import ABC, abstractmethod
from functools import wraps
from typing import TYPE_CHECKING, Callable, Iterable, Self, Sequence, Type, cast

from caseconverter import titlecase
from rich.console import Console, ConsoleOptions, ConsoleRenderable
from rich.markdown import Markdown
from rich.text import Text
from rich.traceback import Traceback

from faebryk.libs.logging import ReprHighlighter
from faebryk.libs.util import md_list

if TYPE_CHECKING:
    from faebryk.core.node import Node


logger = logging.getLogger(__name__)


class UserException(Exception):
    """A user-caused exception."""

    # TODO: Add / refine interface for getting user-facing exception information
    # - Origin?
    # - Title?
    # - Description?
    # - Suggestions?
    # - Help text?
    # __print__?

    def __init__(
        self,
        message: str = "",
        *args,
        title: str | None = None,
        markdown: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(message, *args, **kwargs)
        self.message = message
        self._title = title
        self.markdown = markdown

    @property
    def title(self):
        """Return the name of this error, without the "User" prefix."""
        if self._title is not None:
            return self._title

        error_name = self.__class__.__name__
        return titlecase(error_name.removeprefix("User"))

    def get_frozen(self) -> tuple:
        """Return a frozen version of this error."""
        return (self.__class__, self.message, self._title)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> list[ConsoleRenderable]:
        def _markdown(message: str) -> ConsoleRenderable:
            if console.is_terminal:
                return Markdown(message)
            else:
                return Text(message)

        renderables: list[ConsoleRenderable] = []
        if self.title:
            renderables += [Text(self.title, style="bold")]

        renderables += [
            _markdown(self.message)
            if self.markdown
            else ReprHighlighter()(self.message)
        ]

        return renderables


class DeprecatedException(UserException):
    """This feature is deprecated and will be removed in a future version."""


class UserResourceException(UserException):
    """Indicates an issue with a user-facing resource, e.g. layout files."""


class UserDesignCheckException(UserException):
    """Indicates a failing design check"""

    @classmethod
    def from_nodes(cls, message: str, nodes: Sequence["Node"]) -> Self:
        nodes = sorted(nodes, key=lambda n: str(n))
        msg = f"{message}\n{md_list(f'`{node}`' for node in nodes)}"
        return cls(msg)


class Pacman[T: Exception](contextlib.suppress, ABC):
    """
    A yellow spherical object that noms up exceptions.

    Similar to `contextlib.suppress`, but does something with the exception.
    """

    _exceptions: tuple[Type[T], ...]

    def __init__(
        self,
        *exceptions: Type[T],
        default=None,
    ):
        self._exceptions = exceptions
        self.default = default

    @abstractmethod
    def nom_nom_nom(
        self,
        exc: T,
        original_exinfo: tuple[Type[T], T, Traceback],
    ) -> bool | None:
        """
        Do something with the exception.
        Return True if the exception should be raised.
        """
        # return boolean flipped to make default behavior to suppress

    # The following methods are copied and modified from contextlib.suppress
    # type errors are reproduced faithfully

    def _will_be__exit__(self, exctype, excinst, exctb):  # type: ignore  # Faithfully reproduce type error
        # Unlike isinstance and issubclass, CPython exception handling
        # currently only looks at the concrete type hierarchy (ignoring
        # the instance and subclass checking hooks). While Guido considers
        # that a bug rather than a feature, it's a fairly hard one to fix
        # due to various internal implementation details. suppress provides
        # the simpler issubclass based semantics, rather than trying to
        # exactly reproduce the limitations of the CPython interpreter.
        #
        # See http://bugs.python.org/issue12029 for more details
        if exctype is None:
            return
        if issubclass(exctype, self._exceptions):
            return not self.nom_nom_nom(excinst, (exctype, excinst, exctb))  # type: ignore
        if issubclass(exctype, BaseExceptionGroup):
            excinst = cast(BaseExceptionGroup, excinst)
            match, rest = excinst.split(self._exceptions)  # type: ignore
            if self.nom_nom_nom(match, (exctype, match, exctb)):  # type: ignore
                return False
            if rest is None:
                return True
            raise rest
        return False

    # The following methods are copied and modified from contextlib.ContextDecorator

    def _recreate_cm(self):
        """Return a recreated instance of self.

        Allows an otherwise one-shot context manager like
        _GeneratorContextManager to support use as
        a decorator via implicit recreation.

        This is a private interface just for _GeneratorContextManager.
        See issue #11647 for details.
        """
        return self

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwds):
            with self._recreate_cm():
                return func(*args, **kwds)
            return self.default

        return inner


# HACK: this is attached outside the class definition because
# linters/type-checkers are typically smart enough to know about,
# contextlib.supress, but key off it's __exit__ method
# By attaching it here, we don't have static analysis complaining
# that code is unreachable when using Pacman and it's subclasses
# to downgrade exceptions
Pacman.__exit__ = Pacman._will_be__exit__  # type: ignore


def iter_leaf_exceptions[T: Exception](ex: T | BaseExceptionGroup[T]) -> Iterable[T]:
    """Iter through all the non-group exceptions as a dfs pre-order"""
    if isinstance(ex, ExceptionGroup):
        for e in ex.exceptions:
            yield from iter_leaf_exceptions(e)
    else:
        yield cast(T, ex)


class accumulate:
    """
    Collect a group of errors and only raise
    an exception group at the end of execution.
    """

    def __init__(
        self,
        *accumulate_types: Type,
        group_message: str | None = None,
    ) -> None:
        self.errors: list[Exception] = []

        # Set default values for the arguments
        # NOTE: we don't do this in the function signature because
        # we want the defaults to be the same here as in the iter_through_errors
        # function below
        class _Collector(Pacman):
            def nom_nom_nom(s, exc: Exception, original_exinfo) -> None:
                self.errors.extend(iter_leaf_exceptions(exc))

        self.collector = _Collector(*(accumulate_types or (UserException,)))
        self.group_message = group_message or ""

    def collect(self) -> contextlib.suppress:
        return self.collector

    def get_exception(self) -> Exception | None:
        if self.errors:
            # Display unique errors in order
            # FIXME: this is both hard to understand and wildly inefficient
            displayed_errors = []
            for error in self.errors:
                if not any(
                    existing_error.__dict__ == error.__dict__
                    for existing_error in displayed_errors
                ):
                    displayed_errors.append(error)

            if len(displayed_errors) > 1:
                return ExceptionGroup(self.group_message, displayed_errors)
            else:
                return displayed_errors[0]

    def raise_errors(self):
        """
        Raise the collected errors as an exception group.
        """
        if ex := self.get_exception():
            raise ex

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args):
        self.raise_errors()


class downgrade[T: Exception](Pacman):
    """
    Similar to `contextlib.suppress`, but logs the exception instead.
    Can be used both as a context manager and as a function decorator.
    """

    def __init__(
        self,
        *exceptions: Type[T],
        default=None,
        to_level: int = logging.WARNING,
        raise_anyway: bool = False,
        logger: logging.Logger = logger,
    ):
        super().__init__(*exceptions, default=default)

        if to_level >= logging.ERROR:
            raise ValueError("to_level must be less than ERROR")

        self.to_level = to_level
        self.logger = logger
        self.raise_anyway = raise_anyway

    def nom_nom_nom(self, exc: T, original_exinfo):
        if isinstance(exc, ExceptionGroup):
            exceptions = exc.exceptions
        else:
            exceptions = [exc]

        for e in exceptions:
            self.logger.log(self.to_level, str(e), exc_info=e, extra={"markdown": True})

        return self.raise_anyway


class suppress_after_count[T: Exception](Pacman):
    def __init__(
        self,
        limit: int,
        *exceptions: Type[T],
        default: bool = False,
        suppression_warning: str | None = None,
        logger: logging.Logger = logger,
    ):
        super().__init__(*exceptions, default=default)
        self.limit = limit
        self.counter = 0
        self.supression_warning = suppression_warning
        self.logger = logger

    def nom_nom_nom(self, exc: T, original_exinfo):
        self.counter += 1

        if self.counter == self.limit + 1 and self.supression_warning is not None:
            self.logger.warning(self.supression_warning)

        if self.counter <= self.limit:
            return True


def iter_through_errors[T](
    gen: Iterable[T],
    *accumulate_types: Type,
    group_message: str | None = None,
) -> Iterable[tuple[Callable[[], contextlib.suppress], T]]:
    """
    Wraps an iterable and yields:
    - a context manager that collects any ato errors
        raised while processing the iterable
    - the item from the iterable
    """

    with accumulate(*accumulate_types, group_message=group_message) as accumulator:
        for item in gen:
            # NOTE: we don't create a single context manager for the whole generator
            # because generator context managers are a bit special
            yield accumulator.collect, item


@contextlib.contextmanager
def log_user_errors(logger: logging.Logger = logger):
    """
    Log any exceptions raised within the context.
    """
    try:
        yield
    except UserException as e:
        logger.exception(e)
        raise



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from pathlib import Path

import typer

LICENSE = """# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
"""


def update_file(filepath: Path) -> bool:
    """Update a Python file with the license if it's not present."""

    with open(filepath, "r+", encoding="utf-8") as file:
        contents = file.read()

        if LICENSE not in contents:
            # Add the license at the top of the file and keep the rest as is
            file.seek(0, 0)
            file.write(LICENSE + "\n" + contents)
            return True

        return False


def add_license_to_python_files(directory: Path):
    """Add the license to all Python files in a directory that don't have it."""

    if not directory.is_dir():
        typer.echo(f"The provided path '{directory}' is not a directory.")
        raise typer.Exit(code=1)

    any_changed = False
    for filepath in directory.glob("**/*.py"):
        rc = update_file(filepath)
        any_changed = any_changed or rc

    raise typer.Exit(1 if any_changed else 0)


if __name__ == "__main__":
    typer.run(add_license_to_python_files)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import functools
import logging
import operator
from typing import Callable, Iterable, cast

from faebryk.core.parameter import (
    Abs,
    Add,
    GreaterOrEqual,
    GreaterThan,
    Intersection,
    Is,
    IsSubset,
    Log,
    Multiply,
    Not,
    Or,
    Power,
    Round,
    Sin,
    SymmetricDifference,
    Union,
)
from faebryk.core.solver.algorithm import algorithm
from faebryk.core.solver.mutator import Mutator
from faebryk.core.solver.utils import (
    CanonicalExpression,
    MutatorUtils,
    SolverAll,
    SolverLiteral,
    make_lit,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval_Disjoint,
)
from faebryk.libs.sets.sets import BoolSet, P_Set

logger = logging.getLogger(__name__)


def _multi(op: Callable, init=None) -> Callable:
    def wrapped(*args):
        if init is not None:
            init_lit = make_lit(init)
            args = [init_lit, init_lit, *args]
        assert args
        return functools.reduce(op, args)

    return wrapped


# TODO consider making the oprerator property of the expression type

_CanonicalExpressions = {
    Add: _multi(operator.add, 0),
    Multiply: _multi(operator.mul, 1),
    Power: operator.pow,
    Round: round,
    Abs: abs,
    Sin: Quantity_Interval_Disjoint.op_sin,
    Log: Quantity_Interval_Disjoint.op_log,
    Or: _multi(BoolSet.op_or, False),
    Not: BoolSet.op_not,
    Intersection: _multi(operator.and_),
    Union: _multi(operator.or_),
    SymmetricDifference: operator.xor,
    Is: operator.eq,
    GreaterOrEqual: operator.ge,
    GreaterThan: operator.gt,
    IsSubset: P_Set.is_subset_of,
}

# Pure literal folding -----------------------------------------------------------------


def _exec_pure_literal_operands(
    expr_type: type[CanonicalExpression], operands: Iterable[SolverAll]
) -> SolverLiteral | None:
    operands = list(operands)
    if not all(MutatorUtils.is_literal(o) for o in operands):
        return None
    try:
        return _CanonicalExpressions[expr_type](*operands)
    except (ValueError, NotImplementedError, ZeroDivisionError):
        return None


def _exec_pure_literal_expressions(expr: CanonicalExpression) -> SolverLiteral | None:
    return _exec_pure_literal_operands(type(expr), expr.operands)


@algorithm("Fold pure literal expressions", terminal=False)
def fold_pure_literal_expressions(mutator: Mutator):
    exprs = mutator.nodes_of_types(
        tuple(_CanonicalExpressions.keys()), sort_by_depth=True
    )
    exprs = cast(Iterable[CanonicalExpression], exprs)

    for expr in exprs:
        # TODO is this needed?
        if mutator.has_been_mutated(expr) or mutator.is_removed(expr):
            continue

        # if expression is not evaluatable that's fine
        # just means we can't say anything about the result
        result = _exec_pure_literal_expressions(expr)
        if result is None:
            continue
        # type ignore because function sig is not 100% correct
        mutator.utils.alias_is_literal_and_check_predicate_eval(expr, result)  # type: ignore



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from enum import Enum

from more_itertools import first

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class I2C(ModuleInterface):
    scl: F.ElectricLogic
    sda: F.ElectricLogic

    frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(10 * P.kHz, 3.4 * P.MHz),
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    @L.rt_field
    def requires_pulls(self):
        def pred(signal: F.ElectricSignal, bus: set[Node]):
            interface = signal.get_parent_of_type(I2C)

            assert interface in bus

            return (
                len(bus) > 1
                # arbitrarily choose an interface to represent the bus for this check
                and first(sorted(bus, key=lambda n: str(n))) is interface
                # indicates usage
                and signal.line.crosses_pad_boundary()
            )

        return F.requires_pulls(
            self.scl,
            self.sda,
            pred=pred,
            required_resistance=L.Range(
                1 * (1 - 0.1) * P.kohm, 10 * (1 + 0.1) * P.kohm
            ),
        )

    def terminate(self, owner: Module):
        # TODO: https://www.ti.com/lit/an/slva689/slva689.pdf

        self.pull_up_sda = self.sda.pulled.pull(up=True, owner=owner)
        self.pull_up_scl = self.scl.pulled.pull(up=True, owner=owner)

    class SpeedMode(Enum):
        low_speed = 10 * P.khertz
        standard_speed = 100 * P.khertz
        fast_speed = 400 * P.khertz
        high_speed = 3.4 * P.Mhertz

    @staticmethod
    def define_max_frequency_capability(mode: SpeedMode):
        return L.Range(I2C.SpeedMode.low_speed.value, mode.value)

    def __preinit__(self) -> None:
        self.frequency.add(F.is_bus_parameter())
